#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# del2info v0.1.2
# Copyright (C) 2011 Filip Szymański <f.szymanski@hotmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-


import argparse
import codecs
import ctypes
import os.path
import re
import sys
import time

__version__ = "0.1.2"
__description__ = " del2info, use it to analyze Windows Recycle Bin INFO2 and $I?????? files"
__epilog__ = "examples:\n  del2info.py -i INFO2\n  del2info.py -i $IBA4X90 -o log.txt"

uint32 = ctypes.c_uint32.__ctype_le__
uint64 = ctypes.c_uint64.__ctype_le__
utf16_decoder = codecs.getdecoder("utf_16_le")

pattern = re.compile("\$I[A-Z0-9]{6}$|\$I[A-Z0-9]{6}\.\w")
null_t = re.compile("\x00")

drive_letters = ["A", "B", "C", "D", "E", "F", "G", "H", "I",
					"J", "K", "L", "M", "N", "O", "P", "Q", "R",
					"S", "T", "U", "V", "W", "X", "Y", "Z",]


class StdoutToFile(object):

    def __init__(self, file_name):
		self.stdout = sys.stdout
		try:
			self._file = open(file_name, "a")
		except IOError as err:
			parser.exit(2, "del2info: error: %s\n" % err)
	# end def

    def write(self, text):
		self.stdout.write(text)
		self._file.write(text.encode("utf-8"))
	# end def

    def close(self):
		self.stdout.close()
		self._file.close()
	# end def
# end class


def INFO2(file_name):

	"""Windows XP or older (INFO2 ver.5).	"""

	try:
		_file = open(file_name, "rb")
	except IOError as err:
		parser.exit(2, "del2info: error: %s\n" % err)

	data = _file.read(4)
	info2_version = uint32.from_buffer_copy(data).value

	_file.seek(12)
	data = _file.read(4)
	record_size = uint32.from_buffer_copy(data).value

	if info2_version != 5 or record_size != 800:
		parser.exit(1, "del2info: error: invalid INFO2 file\n")
	# end if

	while True:
		try:
			_file.read(4)

			data = _file.read(260)
			null_t_pos = null_t.search(data, 1)
			if null_t_pos != -1:
				_ascii_name_ = re.sub(null_t, "?", data[:null_t_pos.start()])

			if null_t.match(data):
				_on_disk_ = "No"
			else:
				_on_disk_ = "Yes"
			# end if

			data = _file.read(4)
			_record_number_ = uint32.from_buffer_copy(data).value

			data = _file.read(4)
			drive_number = uint32.from_buffer_copy(data).value
			if drive_number <= 25:
				_drive_ = "(%s:)" % drive_letters[drive_number]
			elif drive_number == 26:
				_drive_ = "(\\)"
			else:
				_drive_ = "(?)"
			# end if

			data = _file.read(8)
			_date_time_ = translate_time(data)

			data = _file.read(4)
			_file_size_ = uint32.from_buffer_copy(data).value

			data = _file.read(516)
			utf8_data = utf16_decoder(data, "ignore")[0]
			if utf8_data:
				_unicode_name_ = re.split(null_t, utf8_data, 1)[0]
			else:
				_unicode_name_ = re.split(null_t, data, 1)[0]
			# end if

		except (AttributeError, ValueError):
			break
	# end while

		print "File name (ascii): %s" % _ascii_name_
		print "File name (unicode): %s" % _unicode_name_
		print "File size: %d bytes" % _file_size_
		print "Deleted (date and time): %s" % _date_time_
		print "Index number: %d" % _record_number_
		print "Drive: %s" % _drive_
		print "File on disk: %s\n" % _on_disk_

	_file.close()
# end def


def SI(file_name):

	"""Windows Vista or later.	"""

	try:
		file_size = os.path.getsize(file_name)
	except OSError as err:
		parser.exit(2, "del2info: error: %s\n" % err)

	if file_size != 543 and file_size != 544:
			parser.exit(1, "del2info: error: invalid $I?????? file\n")
	# end if

	try:
		_file = open(file_name, "rb")
	except IOError as err:
		parser.exit(2, "del2info: error: %s\n" % err)

	data = _file.read(8)
	si_version = uint64.from_buffer_copy(data).value
	if si_version != 1:
		parser.exit(1, "del2info: error: invalid $I?????? file\n")
	# end if

	data = _file.read(8)
	_file_size_ = uint64.from_buffer_copy(data).value

	data = _file.read(8)
	_date_time_ = translate_time(data)

	data = _file.read()
	utf8_data = utf16_decoder(data, "ignore")[0]
	if utf8_data:
		_unicode_name_ = re.split(null_t, utf8_data, 1)[0]
	else:
		_unicode_name_ = re.split(null_t, data, 1)[0]
	# end if

	print "File name (unicode): %s" % _unicode_name_
	print "File size: %d bytes" % _file_size_
	print "Deleted (date and time): %s\n" % _date_time_

	_file.close()
# end def


def translate_time(data):
	timestamp = uint64.from_buffer_copy(data).value
	if timestamp & 9223372036854775808:
		parser.exit(1, "del2info: error: invalid timestamp: %d\n" % timestamp)
	# end if

	# I hope to get the correct date and time.
	seconds = (timestamp - 116444736000000000) // 10000000
	return time.ctime(seconds)
# end def


if __name__ == "__main__":
	try:
		parser = argparse.ArgumentParser(prog="del2info", formatter_class=argparse.RawDescriptionHelpFormatter,
											description=__description__, epilog=__epilog__)
		parser.add_argument("-v", "--version", action="version", version="%(prog)s v"+__version__)
		parser.add_argument("-i", dest="index_file", help="Recycle Bin INFO2 or $I?????? file")
		parser.add_argument("-o", dest="output_file", help="create a log file")

		args = parser.parse_args()

		if len(sys.argv) == 1:
			parser.print_help()
		# end if

		if args.output_file:
			sys.stdout = StdoutToFile(args.output_file)
		# end if

		if args.index_file:
			if os.path.basename(args.index_file) == "INFO2":
				INFO2(args.index_file)
			elif re.search(pattern, os.path.basename(args.index_file)):
				SI(args.index_file)
			else:
				parser.exit(1, "del2info: error: invalid file name: INFO2 or $I??????\n")
			# end if
		# end if

	except KeyboardInterrupt:
		parser.exit(0, "\ninterrupted by user ...\n")
# end if


# Home page:
# http://code.google.com/p/pydetective/
